@@ -1,5 +1,5 @@
 /* Definitions for GCC.  Part of the machine description for CRIS.
-   Copyright (C) 1998, 1999, 2000, 2001 Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
    Contributed by Axis Communications.  Written by Hans-Peter Nilsson.
 
 This file is part of GNU CC.
@@ -59,6 +59,13 @@ Boston, MA 02111-1307, USA.  */
 	abort ();						\
     } while (0)
 
+#define LOSE_AND_RETURN(msg, x)			\
+  do						\
+    {						\
+      cris_operand_lossage (msg, x);		\
+      return;					\
+    } while (0)
+
 struct cris_deferred_string
 {
   int labelno;
@@ -86,9 +93,6 @@ static struct cris_deferred_string cris_deferred_strings[STR_HASHSIZE_MAX];
 static struct cris_cdtor *ctors = NULL;
 static struct cris_cdtor *dtors = NULL;
 
-/* Fix for reg_overlap_mentioned_p.  */
-static int cris_reg_overlap_mentioned_p PARAMS ((rtx, rtx));
-
 /* This little fix suppresses the 'u' or 's' when '%e' in assembly
    pattern.  */
 static char cris_output_insn_is_bound = 0;
@@ -99,26 +103,24 @@ static char cris_output_insn_is_bound = 0;
    just the "sym:GOTOFF" part.  */
 static int cris_pic_sympart_only = 0;
 
-static void
-cris_print_base PARAMS ((rtx, FILE *));
+/* Fix for reg_overlap_mentioned_p.  */
+static int cris_reg_overlap_mentioned_p PARAMS ((rtx, rtx));
 
-static void
-cris_print_index PARAMS ((rtx, FILE *));
+static void cris_print_base PARAMS ((rtx, FILE *));
 
-static void
-cris_init_machine_status PARAMS ((struct function *));
+static void cris_print_index PARAMS ((rtx, FILE *));
 
-static int
-cris_initial_frame_pointer_offset PARAMS ((void));
+static void cris_init_machine_status PARAMS ((struct function *));
 
-static void
-cris_save_machine_status PARAMS ((struct function *));
+static int cris_initial_frame_pointer_offset PARAMS ((void));
 
-static void
-cris_restore_machine_status PARAMS ((struct function *));
+static int saved_regs_mentioned PARAMS ((rtx));
 
-static int
-saved_regs_mentioned PARAMS ((rtx));
+static void cris_operand_lossage PARAMS ((const char *, rtx));
+
+static void cris_save_machine_status PARAMS ((struct function *));
+
+static void cris_restore_machine_status PARAMS ((struct function *));
 
 /* The function cris_function_epilogue puts the last insn to output here.
    Used in delay_slots_for_epilogue and function_epilogue.  */
@@ -510,6 +512,20 @@ cris_op_str (x)
   }
 }
 
+/* Emit an error message when we're in an asm, and a fatal error for
+   "normal" insns.  Formatted output isn't easily implemented, since we
+   use output_operand_lossage to output the actual message and handle the
+   categorization of the error.  */
+
+static void
+cris_operand_lossage (msg, op)
+     const char *msg;
+     rtx op;
+{
+  debug_rtx (op);
+  output_operand_lossage (msg);
+}
+
 /* Print an index part of an address to file.  */
 
 static void
@@ -561,7 +577,8 @@ cris_print_index (index, file)
 	fprintf (file, "[$%s].d", reg_names[REGNO (inner)]);
     }
   else
-    fatal_insn ("Unexpected index-type in cris_print_index", index);
+    cris_operand_lossage ("unexpected index-type in cris_print_index",
+			  index);
 }
 
 /* Print a base rtx of an address to file.  */
@@ -576,7 +593,8 @@ cris_print_base (base, file)
   else if (GET_CODE (base) == POST_INC)
     fprintf (file, "$%s+", reg_names[REGNO (XEXP (base, 0))]);
   else
-    fatal_insn ("Unexpected base-type in cris_print_base", base);
+    cris_operand_lossage ("unexpected base-type in cris_print_base",
+			  base);
 }
 
 /* Usable as a guard in expressions.  */
@@ -1102,7 +1120,7 @@ cris_function_epilogue (file, size)
 	      fprintf (file, save_last);
 	      *save_last = 0;
 	    }
-	
+
 	  if (file)
 	    fprintf (file, "\tAdd%s %d,$sp\n",
 		     ADDITIVE_SIZE_MODIFIER (argspace_offset),
@@ -1249,7 +1267,7 @@ cris_print_operand (file, x, code)
 	 and < 0, i.e print 255 or 65535 as -1, 254, 65534 as -2, etc.  */
       if (GET_CODE (x) != CONST_INT
 	  || ! CONST_OK_FOR_LETTER_P (INTVAL (x), 'O'))
-	fatal_insn ("Internal: Invalid operand with 'b'", x);
+	LOSE_AND_RETURN ("invalid operand for 'b' modifier", x);
       fprintf (file, "%d", INTVAL (x)| (INTVAL (x) <= 255 ? ~255 : ~65535));
       return;
 
@@ -1260,8 +1278,8 @@ cris_print_operand (file, x, code)
 
     case 'v':
       /* Print the operand without the PIC register.  */
-      if (! flag_pic || ! cris_gotless_symbol (x))
-	fatal_insn ("Internal: Invalid operand with 'v'", x);
+      if (! flag_pic || ! CONSTANT_P (x) || ! cris_gotless_symbol (x))
+	LOSE_AND_RETURN ("invalid operand for 'v' modifier", x);
       cris_pic_sympart_only++;
       cris_output_addr_const (file, x);
       cris_pic_sympart_only--;
@@ -1270,15 +1288,15 @@ cris_print_operand (file, x, code)
     case 'P':
       /* Print the PIC register.  Applied to a GOT-less PIC symbol for
          sanity.  */
-      if (! flag_pic || ! cris_gotless_symbol (x))
-	fatal_insn ("Internal: Invalid operand with 'P'", x);
+      if (! flag_pic || ! CONSTANT_P (x) || ! cris_gotless_symbol (x))
+	LOSE_AND_RETURN ("invalid operand for 'P' modifier", x);
       fprintf (file, "$%s", reg_names [PIC_OFFSET_TABLE_REGNUM]);
       return;
 
     case 'p':
       /* Adjust a power of two to its log2.  */
       if (GET_CODE (x) != CONST_INT || exact_log2 (INTVAL (x)) < 0 )
-	fatal_insn ("Internal: Invalid operand with 'p'", x);
+	LOSE_AND_RETURN ("invalid operand for 'p' modifier", x);
       fprintf (file, "%d", exact_log2 (INTVAL (x)));
       return;
 
@@ -1317,7 +1335,7 @@ cris_print_operand (file, x, code)
 	 w for -32768 <= x <= 65535, else abort.  */
       if (GET_CODE (x) != CONST_INT
 	  || INTVAL (x) < -32768 || INTVAL (x) > 65535)
-	fatal_insn ("Internal: Invalid operand with 'z'", x);
+	LOSE_AND_RETURN ("invalid operand for 'z' modifier", x);
       putc (INTVAL (x) >= -128 && INTVAL (x) <= 255 ? 'b' : 'w', file);
       return;
 
@@ -1345,13 +1363,13 @@ cris_print_operand (file, x, code)
 	      return;
 	    }
 	  else
-	    fatal_insn ("Internal: Invalid operand with 'H'", x);
+	    LOSE_AND_RETURN ("invalid operand for 'H' modifier", x);
 
 	case REG:
 	  /* Print reg + 1.  Check that there's not an attempt to print
 	     high-parts of registers like stack-pointer or higher.  */
 	  if (REGNO (operand) > STACK_POINTER_REGNUM - 2)
-	    fatal ("Internal: Bad register: %d", REGNO (operand));
+	    LOSE_AND_RETURN ("bad register", operand);
 	  fprintf (file, "$%s", reg_names[REGNO (operand) + 1]);
 	  return;
 
@@ -1374,7 +1392,7 @@ cris_print_operand (file, x, code)
 	  }
 
 	default:
-	  fatal_insn ("Internal: Invalid operand for 'H'", x);
+	  LOSE_AND_RETURN ("invalid operand for 'H' modifier", x);
 	}
 
     case 'L':
@@ -1388,7 +1406,7 @@ cris_print_operand (file, x, code)
       if (GET_CODE (operand) != SIGN_EXTEND
 	  && GET_CODE (operand) != ZERO_EXTEND
 	  && GET_CODE (operand) != CONST_INT)
-	fatal_insn ("Internal: Invalid operand with 'e'", x);
+	LOSE_AND_RETURN ("invalid operand for 'e' modifier", x);
 
       if (cris_output_insn_is_bound)
 	{
@@ -1405,7 +1423,7 @@ cris_print_operand (file, x, code)
       /* Print the size letter of the inner element.  We can do it by
 	 calling ourselves with the 's' modifier.  */
       if (GET_CODE (operand) != SIGN_EXTEND && GET_CODE (operand) != ZERO_EXTEND)
-	fatal_insn ("Internal: Invalid operand with 'm'", x);
+	LOSE_AND_RETURN ("invalid operand for 'm' modifier", x);
       cris_print_operand (file, XEXP (operand, 0), 's');
       return;
 
@@ -1424,7 +1442,7 @@ cris_print_operand (file, x, code)
       /* When emitting an add for the high part of a DImode constant, we
 	 want to use addq for 0 and adds.w for -1.  */
       if (GET_CODE (operand) != CONST_INT)
-	fatal_insn ("Internal: Invalid operand with 'A' output modifier", x);
+	LOSE_AND_RETURN ("invalid operand for 'A' modifier", x);
       fprintf (file, INTVAL (operand) < 0 ? "adds.w" : "addq");
       return;
 
@@ -1432,7 +1450,7 @@ cris_print_operand (file, x, code)
       /* When emitting an sub for the high part of a DImode constant, we
 	 want to use subq for 0 and subs.w for -1.  */
       if (GET_CODE (operand) != CONST_INT)
-	fatal_insn ("Internal: Invalid operand with 'D' output modifier", x);
+	LOSE_AND_RETURN ("invalid operand for 'D' modifier", x);
       fprintf (file, INTVAL (operand) < 0 ? "subs.w" : "subq");
       return;
 
@@ -1446,8 +1464,7 @@ cris_print_operand (file, x, code)
       /* Print the size letter for an operand to a MULT, which must be a
 	 const_int with a suitable value.  */
       if (GET_CODE (operand) != CONST_INT || INTVAL (operand) > 4)
-	fatal_insn ("Internal: Invalid operand with 'T'", x);
-
+	LOSE_AND_RETURN ("invalid operand for 'T' modifier", x);
       fprintf (file, "%s", mults[INTVAL (operand)]);
       return;
 
@@ -1456,12 +1473,7 @@ cris_print_operand (file, x, code)
       break;
 
     default:
-      {
-#define BADFORMAT "Internal: Invalid operand for '%c'"
-	char s[sizeof BADFORMAT];
-	sprintf (s, BADFORMAT, code);
-	fatal_insn (s, x);
-      }
+      LOSE_AND_RETURN ("invalid operand modifier letter", x);
     }
 
   /* Print an operand as without a modifier letter. */
@@ -1517,7 +1529,7 @@ cris_print_operand (file, x, code)
 	if (GET_CODE (reg) != REG
 	    || (GET_CODE (XEXP (operand, 0)) != CONST_INT
 		&& GET_CODE (XEXP (operand, 1)) != CONST_INT))
-	  fatal_insn ("Can't print operand", x);
+	  LOSE_AND_RETURN ("unexpected multiplicative operand", x);
 
 	cris_print_base (reg, file);
 	fprintf (file, ".%c",
@@ -1537,7 +1549,7 @@ cris_print_operand (file, x, code)
 	  return;
 	}
 
-      fatal_insn ("Internal: Cannot decode operand", x);
+      LOSE_AND_RETURN ("unexpected operand", x);
     }
 }
 
@@ -1572,7 +1584,7 @@ cris_print_operand_address (file, x)
 	  cris_print_index (x1, file);
 	}
       else
-	fatal_insn ("Internal: This is not a recognized address", x);
+	LOSE_AND_RETURN ("unrecognized address", x);
     }
   else if (GET_CODE (x) == MEM)
     {
@@ -1582,7 +1594,7 @@ cris_print_operand_address (file, x)
       putc (']', file);
     }
   else
-    fatal_insn ("Internal: This is not a recognized address", x);
+    LOSE_AND_RETURN ("unrecognized address", x);
 
   putc (']', file);
 }
@@ -1935,17 +1947,34 @@ cris_notice_update_cc (exp, insn)
 		{
 		  /* For "move.S rz,[rx=ry+o]" and "clear.S [rx=ry+o]",
 		     say flags are not changed, except for overlap.  */
-		  if ((cc_status.value1
-		       && cris_reg_overlap_mentioned_p (XEXP
-							(XVECEXP
-							 (exp, 0, 0), 0),
-							cc_status.value1))
-		      || (cc_status.value2
-			  && cris_reg_overlap_mentioned_p (XEXP
-							   (XVECEXP
-							    (exp, 0, 1), 0),
-							   cc_status.value2)))
-		    CC_STATUS_INIT;
+		  if (cc_status.value1
+		      && cris_reg_overlap_mentioned_p (XEXP
+						       (XVECEXP
+							(exp, 0, 0), 0),
+						       cc_status.value1))
+		    cc_status.value1 = 0;
+
+		  if (cc_status.value1
+		      && cris_reg_overlap_mentioned_p (XEXP
+						       (XVECEXP
+							(exp, 0, 1), 0),
+						       cc_status.value1))
+		    cc_status.value1 = 0;
+
+		  if (cc_status.value2
+		      && cris_reg_overlap_mentioned_p (XEXP
+						       (XVECEXP
+							(exp, 0, 0), 0),
+						       cc_status.value2))
+		    cc_status.value2 = 0;
+
+		  if (cc_status.value2
+		      && cris_reg_overlap_mentioned_p (XEXP
+						       (XVECEXP
+							(exp, 0, 1), 0),
+						       cc_status.value2))
+		    cc_status.value2 = 0;
+
 		  return;
 		}
 	    }
@@ -2788,7 +2817,7 @@ cris_override_options ()
 	 further errors.  */
       if (! TARGET_LINUX)
 	{
-	  error ("-fPIC only supported together with -mlinux");
+	  error ("-fPIC and -fpic are only supported together with -mlinux");
 	  flag_pic = 0;
 	}
 
@@ -3021,11 +3050,11 @@ cris_output_addr_const (file, x)
 		fprintf (file, ":GOT]");
 	    }
 	  else
-	    fatal_insn ("Unexpected PIC symbol", x);
+	    LOSE_AND_RETURN ("unexpected PIC symbol", x);
 
 	  /* Sanity check.  */
 	  if (! current_function_uses_pic_offset_table)
-	    fatal ("Emitting PIC operand, but PIC register isn't set up");
+	    output_operand_lossage ("PIC register isn't set up");
 	}
       else
 	assemble_name (file, XSTR (x, 0));
@@ -3115,7 +3144,7 @@ cris_output_addr_const (file, x)
       break;
 
     default:
-      fatal_insn ("Unexpected address expression", x);
+      LOSE_AND_RETURN ("unexpected address expression", x);
     }
 }
 
