@@ -1,5 +1,5 @@
 /* Definitions for GCC.  Part of the machine description for CRIS.
-   Copyright (C) 1998, 1999, 2000, 2001 Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
    Contributed by Axis Communications.  Written by Hans-Peter Nilsson.
 
 This file is part of GNU CC.
@@ -205,15 +205,21 @@ extern int cris_next_int_is_cdtor;
    time being.
 
    Note that -melf overrides -maout except that a.out-compiled libraries
-   are linked in (multilibbing).  */
+   are linked in (multilibbing).
+
+   The option -shared implies -mlinux *at link-time* to cope with broken
+   libtools.  */
 #define LINK_SPEC \
  "%{sim2:%{!T*:-Tdata 0x4000000 -Tbss 0x8000000}}\
   %{melinux:-Ur -d %{!shlib:%{!symbolic:-Bstatic}}}\
   %{melinux-stacksize=*:-defsym __Stacksize=%*}\
-  %{!mlinux:%{!melinux:%{!maout|melf: -mcriself}}%{melinux|maout: -mcrisaout}}\
-  %{mlinux:-mcrislinux -rpath-link linux%s \
-    %{shared} %{static}\
-    %{!shared:%{!static:%{rdynamic:-export-dynamic}}}}\
+  %{!mlinux:%{!shared:%{!melinux:%{!maout|melf: -mcriself}}}\
+   %{melinux|maout: -mcrisaout}}\
+  %{mlinux:\
+    %{!shared:-mcrislinux -rpath-link linux%s %{static}\
+     %{!static:%{rdynamic:-export-dynamic}}}}\
+  %{shared:-mcrislinux -shared -rpath-link linux%s %{static}\
+    %{!static:%{rdynamic:-export-dynamic}}}\
   %{O2|O3:%{!maout|melf:%{!melinux: --gc-sections}}}\
   %{symbolic:-Bdynamic} %{shlib:-Bdynamic} %{static:-Bstatic}\
   %{v:--verbose}"
@@ -222,15 +228,18 @@ extern int cris_next_int_is_cdtor;
    libsc.a if -sim is given to the driver.  Repeat -lc -lsysX
    {X=sim,linux}, because libsysX needs (at least) errno from libc, and
    then we want to resolve new unknowns in libc against libsysX, not
-   libnosys.  Assume everything is in libc for -mlinux.  */
+   libnosys.  Assume everything is in libc for -mlinux.
+   The option -shared implies -mlinux at link-time.  */
 #define LIB_SPEC \
  "%{melinux:-lc -lsyslinux -lc -lsyslinux -lic}\
   %{mlinux:%{!shared:%{pthread:-lpthread}} -lc}\
+  %{shared:%{!mlinux:-lc}}\
   %{!melinux:\
-   %{!mlinux:%{sim*:-lc -lsyssim -lc -lsyssim}\
-    %{!sim*:%{g*:-lg}\
-      %{!p:%{!pg:-lc}}%{p:-lc_p}%{pg:-lc_p} -lbsp}\
-    -lnosys}}"
+   %{!mlinux:\
+    %{!shared:%{sim*:-lc -lsyssim -lc -lsyssim}\
+     %{!sim*:%{g*:-lg}\
+       %{!p:%{!pg:-lc}}%{p:-lc_p}%{pg:-lc_p} -lbsp}\
+     -lnosys}}}"
 
 /* Linker startfile options; crt0 flavors.
 
@@ -245,26 +254,28 @@ extern int cris_next_int_is_cdtor;
    -melinux, and crtbegin2.o then only unless -mlinux.  The file crtbegin2
    corresponds to crti for linux, but handles the combinations of
    linking ELF->ELF, ELF->a.out and a.out->a.out.  Note that crti and crtn
-   for GNU/Linux are always compiled -fpic.  */
+   for GNU/Linux are always compiled -fpic.  Remember that -shared
+   implies -mlinux.  */
 #define STARTFILE_SPEC \
- "%{melinux:%{!mlinux:crt0.o%s}}\
-  %{mlinux:\
-   %{!shared:crt1.o%s crti.o%s crtbegin.o%s}\
-   %{shared:crti.o%s crtbeginS.o%s}}\
+ "%{melinux:%{!mlinux:%{!shared:crt0.o%s}}}\
+  %{mlinux:%{!shared:crt1.o%s crti.o%s crtbegin.o%s}}\
+  %{shared:crti.o%s crtbeginS.o%s}\
   %{!melinux:\
-   %{!mlinux:crtbegin2.o%s crtbegin.o%s\
-    %{sim2:s2crt0.o%s}\
-    %{!sim2:%{sim:scrt0.o%s}\
-     %{!sim:%{pg:gcrt0.o%s}\
-      %{!pg:%{p:mcrt0.o%s}%{!p:crt0.o%s}}}}}}"
+   %{!mlinux:\
+    %{!shared:crtbegin2.o%s crtbegin.o%s\
+     %{sim2:s2crt0.o%s}\
+     %{!sim2:%{sim:scrt0.o%s}\
+      %{!sim:%{pg:gcrt0.o%s}\
+       %{!pg:%{p:mcrt0.o%s}%{!p:crt0.o%s}}}}}}}"
 
 /* Remove elfos.h definition.  */
 #undef ENDFILE_SPEC
 #define ENDFILE_SPEC \
  "%{melinux:%{melf:crtend.o%s}}\
   %{!melinux:\
-   %{mlinux:%{shared:crtendS.o%s crtn.o%s}%{!shared:crtend.o%s crtn.o%s}}\
-   %{!mlinux:crtend.o%s}}"
+   %{mlinux:%{!shared:crtend.o%s crtn.o%s}}\
+   %{shared:crtendS.o%s crtn.o%s}\
+   %{!mlinux:%{!shared:crtend.o%s}}}"
 
 
 /* Node: Run-time Target */
@@ -495,7 +506,7 @@ extern int target_flags;
    with the base gcc version among users, when we ask which version of
    gcc-cris they are using.  Please use some flavor of "R<number>" for
    the version (no need for major.minor versions, I believe).  */
-#define TARGET_VERSION fprintf (stderr, " [CRIS, Axis GCC release R36]");
+#define TARGET_VERSION fprintf (stderr, " [CRIS, Axis GCC release R41]");
 
 #define OVERRIDE_OPTIONS cris_override_options ()
 
@@ -842,16 +853,13 @@ enum reg_class {NO_REGS, ALL_REGS, LIM_REG_CLASSES};
   || CONSTANT_INDEX_P (X)					\
  )
 
-/* FIXME: Bug below: We can't have XEXP (X, 0)) both be MEM and a
-   CONSTANT_P.  Parens don't match indentation.  */
-
 #define EXTRA_CONSTRAINT_T(X)						\
  (									\
-  /* Three-address-operands.  All are indirect-memory:  */		\
+  /* Memory three-address operand.  All are indirect-memory:  */	\
   GET_CODE (X) == MEM							\
   && ((GET_CODE (XEXP (X, 0)) == MEM					\
-  /* Double indirect: [[reg]] or [[reg+]]?  */				\
-  && (BASE_OR_AUTOINCR_P (XEXP (XEXP (X, 0), 0))))			\
+       /* Double indirect: [[reg]] or [[reg+]]?  */			\
+       && (BASE_OR_AUTOINCR_P (XEXP (XEXP (X, 0), 0))))			\
       /* Just an explicite indirect reference: [const]?  */		\
       || CONSTANT_P (XEXP (X, 0))					\
       /* Something that is indexed; [...+...]?  */			\
@@ -1351,21 +1359,21 @@ struct cum_args {int regs;};
 									\
 	  if (REGNO (XEXP (X, 1)) >= FIRST_PSEUDO_REGISTER)		\
 	    {								\
-	      /* Second reg is pseudo, reload it. */			\
+	      /* Second reg is pseudo, reload it.  */			\
 	      push_reload (XEXP (X, 1), NULL_RTX, &XEXP (X, 1), 	\
-			   NULL_PTR,					\
+			   NULL,					\
 			   GENERAL_REGS, GET_MODE (X), VOIDmode, 0, 0,	\
 			   OPNUM, TYPE);				\
 	      something_reloaded = 1;					\
 	    }								\
 									\
-	  if (REG_P (XEXP (XEXP (X, 0), 0))				\
-	      && (REGNO (XEXP (XEXP (X, 0), 0))				\
+	  if (REG_P (XEXP (XEXP (XEXP (X, 0), 0), 0))			\
+	      && (REGNO (XEXP (XEXP (XEXP (X, 0), 0), 0))		\
 		  >= FIRST_PSEUDO_REGISTER))				\
 	    {								\
-	      /* First one is a pseudo - reload that. */		\
-	      push_reload (XEXP (XEXP (X, 0), 0), NULL_RTX,		\
-			   &XEXP (XEXP (X, 0), 0), NULL_PTR, 		\
+	      /* First one is a pseudo - reload that.  */		\
+	      push_reload (XEXP (XEXP (XEXP (X, 0), 0), 0), NULL_RTX,	\
+			   &XEXP (XEXP (XEXP (X, 0), 0), 0), NULL, 	\
 			   GENERAL_REGS,				\
 			   GET_MODE (X), VOIDmode, 0, 0, OPNUM, TYPE);	\
 	      something_reloaded = 1;					\
@@ -1377,7 +1385,7 @@ struct cum_args {int regs;};
 		      >= FIRST_PSEUDO_REGISTER)))			\
 	    /* Reload the sign_extend.	Happens if neither reg is a	\
 	       pseudo, or the first one was inside post_increment.  */	\
-	    push_reload (XEXP (X, 0), NULL_RTX, &XEXP (X, 0), NULL_PTR,	\
+	    push_reload (XEXP (X, 0), NULL_RTX, &XEXP (X, 0), NULL,	\
 			 GENERAL_REGS, GET_MODE (X), VOIDmode, 0, 0,	\
 			 OPNUM, TYPE);					\
 	  goto WIN;							\
